# -*- coding: utf-8 -*-
# contacts.models
import datetime

import django.contrib.auth.models as auth_models
from django.contrib import messages
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from location_field.models.plain import PlainLocationField
from positions.fields import PositionField

from ..references import models as ref
from ..siret import CheckSiret
from .fields_choices import (
    ACTIVITE_BIO,
    ACTIVITE_LISTE,
    CHOIX_ANNEE,
    CHOIX_CIVILITE,
    CHOIX_MOIS,
    CHOIX_PARCS_NATURELS_REGIONAUX,
    CHOIX_TERME,
    CHOIX_TYPE_CIRCUIT,
    CHOIX_TYPE_RDV,
    CHOIX_VISIBILITE,
    MODE_REGLT,
    MTX_TYPE_RDV,
    POND_CONV,
    STATUT_BIO,
    TYPE_AGREMENT,
    TYPE_CIRCUIT,
    TYPE_PRODUIT,
    TYPE_TRANS,
)
from .model_validator import validate_plain_location
from .settings_messages import AVERTISSEMENT_MESSAGE as w_msg
from .settings_messages import ERROR_MESSAGE as e_msg
from .tables import (
    csv_assolement,
    csv_commercialisation,
    csv_donneesbio,
    csv_personne,
    csv_projetbio,
    csv_rdv,
    csv_rdvpersonne,
    csv_structure,
    csv_transformation,
)

# from functools import partial


LOCAL_DATE_FORMAT = '%d/%m/%Y'


# ## Utilitaires
def get_current_year():
    return datetime.date.today().year


def get_rem_user():
    return auth_models.User.objects.get_or_create(username='deleted')[0]


def no_attrib_error(orig_func):
    "return empty unicode if attrib does not exist"

    def wrapper(*args, **kwargs):
        try:
            return orig_func(*args, **kwargs)
        except AttributeError:
            return ""

    return wrapper


def _field_fb(self, field, field_cond=None, get_other=None):
    "return a field if present else call get_other"
    pcache = '_fb_' + field
    field_cond = field_cond or field
    if not hasattr(self, pcache):
        if getattr(self, field_cond, None):
            itm = getattr(self, field)
        elif get_other:
            itm = get_other()
        else:
            itm = None
        if callable(itm):
            itm = itm()
        setattr(self, pcache, itm)
    return getattr(self, pcache)


def _field_fb_struct(self, field, field_cond=None):
    "return person's field of if empty, structure's one"

    def get_other():
        return self.structure_id and getattr(self.structure, field)

    return _field_fb(self, field, field_cond, get_other)


# ## Classes principales
class AdhManager(auth_models.UserManager):
    use_in_migrations = False

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(groups__isnull=True, is_superuser=False)
        )


class AdhLogin(auth_models.User):
    class Meta:
        proxy = True
        verbose_name = "Compte adhérent"
        verbose_name_plural = "      Comptes adhérents"

    objects_adh = AdhManager()
    objects = auth_models.UserManager()

    def is_adherent(self):
        "poor man test: si ne fait partie de aucun groupe, c'est un adherent"
        return bool(self.groups)


class StructureManager(models.Manager):
    def get_by_natural_key(self, designation):
        return self.get(designation=designation)


class Structure(models.Model):
    "Persones Morales"

    class Meta:
        ordering = ['designation']
        verbose_name = "Personne Morale"
        verbose_name_plural = "         Personnes Morales"
        indexes = [
            models.Index(
                fields=['designation'], name='contact_str_designation_idx'
            )
        ]

    objects = StructureManager()
    attrs_export = csv_structure.attrs_export
    attrs_publipostage = csv_structure.attrs_publipostage
    designation = models.CharField("désignation", max_length=250, unique=True)
    utilisateur_derniere_modification = models.ForeignKey(
        auth_models.User,
        editable=False,
        on_delete=models.PROTECT,
        verbose_name="dernière modification par",
        related_name='adherent_structure',
    )
    adherent = models.ForeignKey(
        AdhLogin,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="compte adhérent",
        help_text="compte permettant la connexion à l'application",
    )
    f_juridique = models.ForeignKey(
        ref.FormeJuridique,
        null=True,
        on_delete=models.PROTECT,
        verbose_name="Forme juridique",
    )
    statut = models.ForeignKey(
        ref.StatutStructure, null=True, blank=True, on_delete=models.SET_NULL
    )
    cat_act_agriculteur = models.BooleanField(
        "agriculteur", blank=True, default=False
    )
    cat_act_collecteur = models.BooleanField(
        "collecteur", blank=True, default=False
    )
    cat_act_transformateur = models.BooleanField(
        "transformateur", blank=True, default=False
    )
    cat_act_commerce_gros = models.BooleanField(
        "com. gros", blank=True, default=False
    )
    cat_act_commerce_detail = models.BooleanField(
        "com. detail", blank=True, default=False
    )
    cat_act_collectivite = models.BooleanField(
        "Collectivité", blank=True, default=False
    )
    cat_act_autre = models.CharField(
        "autre", max_length=255, blank=True, default=''
    )
    url_old_ref = models.URLField(
        "URL ancienne référence",
        max_length=250,
        null=True,
        blank=True,
        editable=False,
    )
    adresse = models.TextField(blank=True, default='')
    commune = models.ForeignKey(
        ref.Commune, null=True, blank=True, on_delete=models.PROTECT
    )
    non_courrier = models.BooleanField(
        "ne pas envoyer de courrier à cette adresse", blank=True, default=False
    )
    # Certaines communes peuvent avoir plusieurs codes postaux
    code_postal = models.CharField(
        max_length=10,
        blank=True,
        default='',
        help_text="Laisser vide pour utiliser celui associé à la commune",
    )
    # la ville contient également le code cedex (à utilser pour l'adresse)
    ville = models.CharField(
        max_length=255,
        blank=True,
        default='',
        help_text="Nom de ville avec code cédex si nécessaire, "
        " si différent du nom de la commune.",
    )
    geolocalisation = PlainLocationField(
        based_fields=['code_postal', 'ville'],
        default='',
        blank=True,
        help_text=(
            "Valeur initiée via le champ « ville ». "
            "Déplacez le marqueur pour affiner le résultat."
            "Zoomez avez la molette de la souris."
        ),
        validators=[validate_plain_location],
    )
    tel_fixe = models.CharField(
        "téléphone fixe", max_length=100, blank=True, default=''
    )
    tel_mobile = models.CharField(
        "téléphone mobile", max_length=100, blank=True, default=''
    )
    fax = models.CharField(max_length=100, blank=True, default='')
    courriel = models.CharField(max_length=250, blank=True, default='')
    siret = models.CharField(
        "N. Siret",
        max_length=14,
        blank=True,
        db_index=True,
        default='',
        help_text="14 chiffres sans espaces",
    )
    pas_de_courriel = models.BooleanField(
        "pas d'envoi automatique de courriel",
        default=False,
        help_text="Cochez cette case si cette structure ne souhaite"
        " pas recevoir de courriel envoyé automatiquement.",
    )
    site_web = models.URLField(
        "URL du site web", max_length=250, null=True, blank=True
    )
    date_creation = models.DateField(
        "date de saisie de l'enregistrement", editable=False
    )
    date_derniere_modification = models.DateField(
        "date de dernière modification des informations", editable=False
    )
    origine_contact = models.CharField(
        "origine du contact", max_length=255, blank=True, default=''
    )
    annee_contact = models.IntegerField(
        "Année du premier contact", null=True, blank=True, choices=CHOIX_ANNEE
    )
    mois_contact = models.IntegerField(
        "Mois du premier contact", null=True, blank=True, choices=CHOIX_MOIS
    )
    sage = models.CharField("SAGE", max_length=255, blank=True, default='')
    commentaire = models.TextField(blank=True, default='')
    etiquettes = models.ManyToManyField(
        ref.Etiquette, blank=True, verbose_name='étiquettes'
    )
    statutbio = models.IntegerField(
        "Statut bio", default=0, choices=STATUT_BIO
    )
    # from agriculteur
    visibilite = models.CharField(
        max_length=255,
        choices=CHOIX_VISIBILITE,
        default=CHOIX_VISIBILITE[0][0],  # personne
        verbose_name="publication sur l'annuaire du site",
    )
    description_activite = models.TextField(blank=True, default='')
    description_activite_map = models.TextField(blank=True, default='')
    description_web = models.TextField(blank=True, default='')
    description_photo = models.ImageField(
        upload_to='photo', verbose_name="photo", blank=True
    )
    description_categorie = models.ManyToManyField(
        ref.DescCategorieStructure, verbose_name="Catégorie"
    )
    description_typevente = models.ManyToManyField(
        ref.DescTypeVenteStructure, verbose_name="Type de vente"
    )
    # si commune.enjeu_eau
    sau_eau = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        null=True,
        blank=True,
        verbose_name="SAU à enjeu eau",
        help_text='Surface agricole utile sur commune à enjeu eau (en ha)',
    )
    sau_eau_bio = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        null=True,
        blank=True,
        verbose_name="SAU en biologique à enjeu eau",
        help_text="Surface agricole utile en biologique"
        " sur commune à enjeu eau (en ha)",
    )
    # fin si commune.enjeu_eau
    parc_naturel_regional = models.CharField(
        max_length=200,
        default='',
        verbose_name="parc naturel régional",
        choices=CHOIX_PARCS_NATURELS_REGIONAUX,
        blank=True,
    )
    # from agriculteurbio
    premiere_adhesion = models.ForeignKey(
        'Adhesion',
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name='premieres_adhesions',
        verbose_name="Première adhésion",
    )
    renouvellement_adhesion = models.ForeignKey(
        'Adhesion',
        null=True,
        blank=True,
        related_name='renouvellements',
        on_delete=models.SET_NULL,
        verbose_name="dernier renouvellement d'adhésion",
    )

    def natural_key(self):
        return [self.designation]

    def url_old_ref_link(self):
        if not self.url_old_ref:
            return '-'
        return format_html('<a href="{0}">{0}</a>', self.url_old_ref)

    url_old_ref_link.short_description = "URL ancienne référence"

    def _get_activite(self):
        act_pri = None
        act_pri_bio = None
        act_annexe = []
        for elm in self.activitestructure_set.all():
            if act_pri and elm.activite_bio < 2:
                act_annexe.append(elm)
            if act_pri_bio and elm.activite_bio == 2:
                act_annexe.append(elm)
            if act_pri is None:
                act_pri = elm
            if act_pri_bio is None and elm.activite_bio == 2:
                act_pri_bio = elm
        self._act_pri = act_pri
        self._act_pri_bio = act_pri_bio
        self._act_annexe = act_annexe

    def _get_activite_attr(self, attr):
        if not hasattr(self, attr):
            self._get_activite()
        return getattr(self, attr)

    def activite_principale(self):
        return self._get_activite_attr('_act_pri')

    activite_principale.short_description = "Activité principale"

    def activite_principale_bio(self):
        return self._get_activite_attr('_act_pri_bio')

    activite_principale_bio.short_description = "Activité principale bio"

    def activites_annexes_list(self):
        return self._get_activite_attr('_act_annexe')

    def activites_annexes_bio_list(self):
        return [
            elm
            for elm in self._get_activite_attr('_act_annexe')
            if elm.activite_bio == 2
        ]

    def activites_annexes_csv(self):
        act_list = self._get_activite_attr('_act_annexe')
        return ", ".join([str(elm) for elm in act_list])

    activites_annexes_csv.short_description = "Activité(s) annexe(s)"

    def activites_annexes(self):
        act_list = self._get_activite_attr('_act_annexe')
        return format_html(
            "<ul>{}</ul>".format(
                "".join(["<li>{}</li>".format(elm) for elm in act_list])
            )
        )

    activites_annexes.short_description = "Activité(s) annexe(s)"

    def cat_acteur(self):
        return "/".join(
            [
                elm
                for elm in (
                    'agriculteur',
                    'collecteur',
                    'transformateur',
                    'commerce_gros',
                    'commerce_detail',
                    'collectivite',
                )
                if getattr(self, 'cat_act_{}'.format(elm))
            ]
        )

    cat_acteur.short_description = "Catégorie d'acteur"

    def __str__(self):
        "Retourne une représentation unicode de la structure."
        return self.designation

    def get_absolute_url(self):
        "Retourne l'URL associée à l'objet."
        return reverse(
            'admin:{0}_{1}_change'.format(
                self._meta.app_label, self.__class__.__name__.lower()
            ),
            args=[self.pk],
        )

    def liste_contacts_associes(self):
        """
        Surcharge de la méthode de Contact.
        Retourne la liste des contacts associés à cet objet. Il
        s'agit ici des personnes associées à la structure.
        """
        return self.personne_set.all()

    def adresse_electronique(self):
        """
        Retourne une liste de 2 éléments:
        - le nom de la structure
        - l'adresse email associée
        Retourne une liste vide si aucune adresse courriel n'est renseignée.
        """
        if not self.courriel:
            return ()
        return (str(self), self.courriel)

    def get_courriels(self):
        """
        Retourne le courriel de la structure et des personnes associées
        (sauf s'ils ne souhaitent pas recevoir d'email automatique)
        """
        courriels = []
        if self.courriel:
            courriels.append(self.adresse_electronique())
        for p in self.liste_contacts_associes():
            courriels.extend(p.get_courriels())
        return courriels

    def clean(self):
        self.set_default_ville_code_postal()
        self.clean_siret()

    def clean_siret(self):
        if self.siret:
            status_siret = CheckSiret(self.siret)
            if status_siret.is_siret:
                self.siret = status_siret.get_verified_value()
            else:
                raise ValidationError(
                    {
                        'siret': ValidationError(
                            "Num. de siret incorrect", code='siret_invalid'
                        )
                    }
                )
            # check unique
            base_qs = self._meta.model.objects.filter(siret=self.siret)
            if self.id:
                base_qs = base_qs.exclude(id=self.id)
            if base_qs.exists():
                raise ValidationError(
                    {
                        'siret': ValidationError(
                            "Num. de siret existant", code='siret_dup'
                        )
                    }
                )

    def set_default_ville_code_postal(self):
        """
        Affecte des valeurs par défaut au code postal et
        à la ville lorsque seule la commune est précisée.
        """
        if self.commune:
            if not self.code_postal:
                self.code_postal = self.commune.code_postal
            if not self.ville:
                self.ville = self.commune.nom

    def save(self, *args, **kwargs):
        """
        Remplit les champs date_creation et date_derniere_modification
        manuellement (la fonctionnalité auto_now de django est douteuse,
        mieux vaut le gérer manuellement).
        """
        rawsave = kwargs.pop('rawsave', False)
        if rawsave:
            models.Model.save(self, *args, **kwargs)
        if self.date_creation is None:
            self.date_creation = datetime.date.today()
        self.date_derniere_modification = datetime.date.today()
        models.Model.save(self, *args, **kwargs)

    def get_date_premiere_adhesion(self):
        """
        Retourne la date de première adhésion, ou None s'il n'est pas abonné.
        """
        try:
            return self.premiere_adhesion.date
        except Adhesion.DoesNotExist:
            return None

    get_date_premiere_adhesion.short_description = "date de première adhésion"

    def get_date_derniere_adhesion(self):
        """
        Retourne la date de dernière adhésion, ou None s'il n'est pas abonné.
        """
        try:
            return self.adhesion_set.latest().date
        except Adhesion.DoesNotExist:
            return None

    get_date_derniere_adhesion.short_description = "date de dernière adhésion"

    def get_date_premiere_conversion(self):
        """
        Retourne la date de première conversion bio, ou None s'il n'y en a pas
        (ce qui ne devrait pas être le cas pour un agriculteur bio, mais
        restons parano^Wprudent).
        """
        conv = self.projetbio_set.filter(
            date_engagement_oc__isnull=False
        ).order_by('date_engagement_oc')
        if len(conv) == 0:
            return None
        return conv[0].date_engagement_oc

    get_date_premiere_conversion.verbose_name = "date de première conversion"
    get_date_premiere_conversion.short_description = (
        "date de première conversion"
    )

    def premiere_conversion(self):
        return self.get_date_premiere_conversion()

    def last_donneesbio(self):
        if not hasattr(self, '_last_dbio'):
            try:
                self._last_dbio = self.donneesbio_set.order_by('-annee')[:1][0]
            except IndexError:
                self._last_dbio = None
        return self._last_dbio

    last_donneesbio.target_model = 'DonneesBio'

    def ass_total_bio(self):
        ldb = self.last_donneesbio()
        return ldb and ldb.ass_total_bio

    ass_total_bio.short_description = "Assolement total (en Ha)"

    def ass_total(self):
        ldb = self.last_donneesbio()
        return ldb and ldb.ass_total

    ass_total.short_description = "Assolement total bio (en Ha)"

    def get_liste_donneesbio(self):
        """
        Retourne la listes des données de l'observatoire de l'agriculteur
        (utilisé pour afficher les liens vers ceux-ci dans la vue de
        modification de l'agriculteur).
        """
        return [
            {'url': dbio.get_absolute_url(), 'annee': dbio.annee}
            for dbio in self.donneesbio_set.order_by('-annee')
        ]

    def get_url_cree_donneesbio(self):
        """
        Retourne l'URL de la vue permettant de créer un nouvel objet DonneesBio
        pour cet agriculteur et d'afficher sa page de modification.
        """
        return reverse('admin:cree_donneesbio', args=[self.id])

    def get_youngest_person(self):
        "return youngest person attached to this structure if any"
        try:
            return self.personne_set.latest('date_naissance')
        except models.ObjectDoesNotExist:
            return None

    get_youngest_person.target_model = 'Personne'

    def get_rdv_coll(self):
        rdv_liste = []
        rdvP_liste = RdvPersonne.objects.filter(
            personne__structure_id=self.id
        ).order_by('-rdv__date')[:10]
        return rdvP_liste
        for rdvP in rdvP_liste:
            if rdvP.rdv not in rdv_liste:
                rdv_liste.append(rdvP.rdv)
        return rdv_liste

    def get_rdv_ind(self):
        return self.rdv_ind_set.all()

    def get_messages(self):
        "fill info/warning messages at save"
        mlist = []
        if self.statutbio >= 1 and not self.personne_set.exists():
            pass
            mlist.append(
                (messages.WARNING, w_msg['message_personne_manquante'])
            )
        if self.statutbio >= 2:
            if self.projetbio_set.exists():
                if not self.projetbio_set.filter(
                    date_engagement_oc__isnull=False
                ).exists():
                    mlist.append(
                        (
                            messages.WARNING,
                            w_msg['message_projet_bio_manquant'],
                        )
                    )
            else:
                mlist.append(
                    (messages.WARNING, w_msg['message_projet_bio_manquant'])
                )
        if self.statutbio >= 1 and not (
            self.ass_total() or self.ass_total_bio()
        ):
            mlist.append((messages.WARNING, w_msg['sau_vide']))
        return mlist


class ActiviteStructureManager(models.Manager):
    def get_by_natural_key(self, structure_nom, nom, activite_bio):
        return self.get(
            structure__designation=structure_nom,
            nom=nom,
            activite_bio=activite_bio,
        )


class ActiviteStructure(models.Model):
    "Sous table Activite pour les Structure"

    class Meta:
        ordering = ('position',)
        unique_together = [['structure', 'nom', 'activite_bio']]

    objects = ActiviteStructureManager()

    structure = models.ForeignKey(Structure, on_delete=models.CASCADE)
    nom = models.CharField(max_length=16, choices=ACTIVITE_LISTE, default='')
    precision = models.CharField(max_length=128, blank=True, default='')
    activite_bio = models.IntegerField(
        "Activité bio ?", default=0, choices=ACTIVITE_BIO
    )
    position = PositionField(
        collection='structure',
        help_text="La 1ère position (0) est l'activité principale, "
        "la 1ère activité bio est l'activité principale bio."
        " La dernière postition est (-1)",
    )

    def natural_key(self):
        return self.structure.natural_key() + [self.nom, self.activite_bio]

    def clean(self):
        if self.nom == 'autre' and not self.precision:  # ni nom ni precision
            raise ValidationError(
                {
                    'precision': ValidationError(
                        "Doit être précisé si le champs 'nom' est vide",
                        code='mandatory_field',
                    )
                }
            )

    def __str__(self):
        """
        [X] NOM
        avec X parmi C (conventionel), P (projet bio), B (bio)
        """
        precision = "({})".format(self.precision) if self.precision else ""
        nom = "autre" if self.nom == 'autre' else self.get_nom_display()
        bio_flag = "CPB"[self.activite_bio]
        return "[{}] {}{}".format(bio_flag, nom, precision)


class ProjetBio(models.Model):
    """
    Projet de passage en bio d'un agriculteur.
    Chaque agriculteur peut en avoir plusieurs, à conserver même après le
    passage en Bio.
    """

    class Meta:
        verbose_name = "projet de passage en bio"
        verbose_name_plural = "      Projets de passage en bio"

    attrs_export = csv_projetbio.attrs_export

    structure = models.ForeignKey(
        'Structure', verbose_name="Personne morale", on_delete=models.PROTECT
    )
    technicien = models.ForeignKey(
        auth_models.User,
        null=True,
        blank=True,
        verbose_name="Suivi par",
        on_delete=models.SET_NULL,
    )
    description = models.TextField(
        blank=True, verbose_name="description", default=''
    )
    freins_identifies = models.TextField(
        "Points favorables et freins à la conversion", blank=True, default=''
    )

    pond_savoirfaire = models.IntegerField(
        "Savoir faire",
        default=0,
        choices=POND_CONV,
        help_text="Manque de savoir-faire ou de connaissances (projet mal "
        "défini...), peur de ne pas être à la hauteur, manque de rigueur, "
        "manque de qualification de l'exploitant, formation initiale non "
        "adaptée",
    )
    pond_motivation = models.IntegerField(
        "Motivation",
        default=0,
        choices=POND_CONV,
        help_text="Manque de motivation, faible sensibilité à la bio, peu "
        "enclin au changement, proche de la retraite, pas envie de réorienter"
        " un système intensif, pas de frein identifié, mauvaise image de la "
        "bio, peur de perte de revenu non-avérée.",
    )
    pond_entourage = models.IntegerField(
        "Entourage",
        default=0,
        choices=POND_CONV,
        help_text=""
        "Pression de l'entourage (familles, conseillers techniques ou "
        "financier, système intégré...)",
    )
    pond_foncier = models.IntegerField(
        "Foncier",
        default=0,
        choices=POND_CONV,
        help_text=""
        "Problème de foncier (manque de surface, menace de perdre des "
        "surfaces, manque de surface accessible aux vaches, qualité des "
        "terres...), pas de foncier dans le cas d'une installation",
    )
    pond_materiel = models.IntegerField(
        "Matériel",
        default=0,
        choices=POND_CONV,
        help_text=""
        "Moyens de productions faibles ou inadaptés et difficilement "
        "adaptables, génétique animale ou végétale peu adaptée, race de "
        "vache, variété de pomme, manque de main d'œuvre, bâtiment inadapté"
        " trop petit ou pas aux normes, bâtiment de stockage inadapté",
    )
    pond_reglementation = models.IntegerField(
        "Réglementation",
        default=0,
        choices=POND_CONV,
        help_text=""
        "Problème réglementaires, mixité, engagé dans un cahier des charges,"
        " forfait fiscal, règle ICPE, MAEt ou PAE.",
    )
    pond_commercialisation = models.IntegerField(
        "Commercialisation",
        default=0,
        choices=POND_CONV,
        help_text=""
        "Commercialisation, difficulté à passer des hausses de prix en vente"
        " directe, produit sans valorisation bio ou sans filière longue "
        "(betterave sucrières, moutons, volaille de chair), plus value jugée"
        " faible en bio, perte de revenu, cahier des charges privé",
    )
    pond_financement = models.IntegerField(
        "Financement",
        default=0,
        choices=POND_CONV,
        help_text=""
        "Manque de financement du projet, manque d'aides financières (Peu ou"
        "pas d'aide à la conversion), dans le cas des installations : peu ou"
        " pas d'apport de capital/incapacité à investir/pas d'accès aux aides"
        " installation",
    )
    pond_maindoeuvre = models.IntegerField(
        "Main d'Oeuvre ",
        default=0,
        choices=POND_CONV,
        help_text=""
        "Manque de qualification de la main d'œuvre à embaucher, Manque de "
        "main d'œuvre",
    )
    suites_a_donner = models.TextField(
        blank=True, verbose_name="suites à donner", default=''
    )
    remarques = models.TextField(
        blank=True, verbose_name="remarques", default=''
    )
    etude_demandee = models.NullBooleanField("Étude demandée ?")
    date_etude = models.DateField("Date de l'étude", null=True, blank=True)
    date_probable_certification = models.CharField(
        "date probable de début de conversion",
        blank=True,
        default='',
        max_length=20,
        choices=CHOIX_TERME,
        help_text="Si vous "
        "choisissez l'option « Date précisée ci-après », merci de préciser"
        " la date.",
    )
    date_certification = models.DateField(
        "Date de début de conversion prévue",
        null=True,
        blank=True,
        help_text="à remplir si vous avez choisi l'option \"Date précisée\"",
    )
    engagement_surf = models.DecimalField(
        "Surface engagée (en ha)",
        null=True,
        blank=True,
        decimal_places=4,
        max_digits=10,
        help_text="utilisez . comme séparateur décimal",
    )
    engagement_oc_enjeu_eau = models.DecimalField(
        "Surface engagée sur commune à enjeu eau (ha)",
        null=True,
        blank=True,
        decimal_places=4,
        max_digits=10,
        help_text="utilisez . comme séparateur décimal",
    )
    installation_bio = models.NullBooleanField("installation bio")
    date_engagement_oc = models.DateField(
        "date d'engagement de l'OC", null=True, blank=True, db_index=True
    )
    commentaire = models.TextField(blank=True, default='')

    def __str__(self):
        """
        Retourne une représentation unicode de l'objet.
        Attention, problème vicieux: l'admin django appelle cette méthode juste
        après la suppression de l'objet dans la base: il ne faut donc pas
        utiliser %d pour self.id car dans ce cas, self.id vaut None et provoque
        une erreur.
        """
        return "projet bio de %s (id #%s)" % (self.structure, self.id)

    def agri_id(self):
        if self.structure:
            return self.structure_id
        return ""

    agri_id.short_description = "ID personne morale"

    def get_messages(self):
        mlist = []
        # check that at least one pond_* is set to '-1'
        res_pond = {}
        for attr in dir(self):
            if attr.startswith('pond_'):
                res_pond.setdefault(getattr(self, attr), 0)
                res_pond[getattr(self, attr)] += 1
        if res_pond.get(-1, 0) == 0:
            mlist.append((messages.WARNING, e_msg['projet_sans_contrainte']))
        if self.installation_bio is None:
            mlist.append(
                (messages.WARNING, e_msg['projet_installation_non_choisi'])
            )
        return mlist

    def get_absolute_url(self):
        """
        Retourne l'URL associée à l'objet.
        """
        return reverse(
            'admin:{0}_{1}_change'.format(
                self._meta.app_label, self.__class__.__name__.lower()
            ),
            args=[self.pk],
        )


class DonneesBioManager(models.Manager):
    def get_by_natural_key(self, structure_nom, annee):
        return self.get(structure__designation=structure_nom, annee=annee)


class DonneesBio(models.Model):
    "Données de l'observatoire pour une année spécifique."

    class Meta:
        verbose_name = "Info annuelle sur l'exploitation"
        verbose_name_plural = "    Infos annuelles sur l'exploitation"
        ordering = ['-annee', 'structure']
        unique_together = [['structure', 'annee']]

    objects = DonneesBioManager()
    attrs_export = csv_donneesbio.attrs_export
    structure = models.ForeignKey(
        'Structure', verbose_name="Agriculteur", on_delete=models.CASCADE
    )
    date_collecte = models.DateField(null=True, blank=True)
    date_modification = models.DateField(null=True, blank=True, editable=False)
    date_validation = models.DateField(null=True, blank=True, editable=False)
    technicien = models.ForeignKey(
        auth_models.User,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        help_text="Personne ayant collecté les données",
    )
    annee = models.IntegerField(
        "année", db_index=True, default=get_current_year
    )
    organisme_certification = models.ForeignKey(
        ref.OrganismeDeCertification,
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name="organisme de certification",
    )
    biodynamie = models.NullBooleanField()
    remarques = models.TextField(
        blank=True, verbose_name="remarques de l'équipe", default=''
    )
    # Aides
    aide_conversion_ab = models.NullBooleanField(
        verbose_name="aide à la conversion à l'agriculture biologique"
    )
    date_aide_conversion_ab = models.DateField(
        'date de cette aide',
        null=True,
        blank=True,
        help_text="En cas d'aide à la conversion, indiquez la date de "
        "signature du contrat (ou de dépôt pour les MAE)",
    )
    aide_maintien_ab = models.NullBooleanField(
        verbose_name="aide au maintien à l'agriculture biologique"
    )
    aide_credit_impot = models.NullBooleanField(
        verbose_name="crédit d'impôt bio"
    )
    aide_certification_bio = models.NullBooleanField(
        verbose_name="aide à la certification bio"
    )
    aide_investissement = models.NullBooleanField(
        verbose_name="aide à l'investissement"
    )
    aide_installation = models.NullBooleanField(
        verbose_name="aide à l'installation"
    )
    aide_mae_t = models.NullBooleanField("MAE T")
    aide_autres = models.CharField(
        "autres mesures, précisez", max_length=255, blank=True, default=''
    )
    aide_accompagnement = models.NullBooleanField(
        "avez-vous besoin d'un accompagnement pour solliciter une aide ?"
    )
    aide_accompagnement_precisions = models.TextField(
        "précisions", blank=True, default=''
    )
    # Main d'oeuvre
    etp_associes = models.DecimalField(
        "ETP d'associés", null=True, blank=True, max_digits=6, decimal_places=2
    )
    etp_familiaux = models.DecimalField(
        "ETP familiaux", null=True, blank=True, max_digits=6, decimal_places=2
    )
    etp_salaries_permanents = models.DecimalField(
        "ETP de salariés permanents",
        null=True,
        blank=True,
        max_digits=6,
        decimal_places=2,
    )
    etp_saisonniers = models.DecimalField(
        "ETP saisonniers",
        null=True,
        blank=True,
        max_digits=6,
        decimal_places=2,
    )
    etp_encadrants = models.DecimalField(
        "ETP encadrants", null=True, blank=True, max_digits=6, decimal_places=2
    )
    personnes_insertion = models.IntegerField(
        "Personnes en contrat d'insertion", null=True, blank=True
    )
    reseaux_engagements = models.TextField(
        "Appartenance à des réseaux et engagements", blank=True, default=''
    )
    # Transmission
    tr_date_fin_prevue = models.DateField(
        "Date de fin d'activité prévue",
        null=True,
        blank=True,
        help_text="donnez une date approximative si possible",
    )
    tr_repreneur_ident = models.NullBooleanField(
        "Repreneur identifié", null=True, blank=True
    )
    tr_precisions = models.TextField(
        "Précisions",
        blank=True,
        default='',
        help_text="Identité du repreneur et autre précisions",
    )
    # Commercialisation
    statut_commercialisation = models.NullBooleanField(
        "utilisation d'un statut spécifique pour la commercialisation"
    )
    statut_choisi = models.ForeignKey(
        ref.StatutCommercialisation,
        null=True,
        blank=True,
        verbose_name="statut choisi",
        help_text="en cas d'utilisation d'un"
        " statut spécifique pour la commercialisation, indiquez-le ici.",
        on_delete=models.PROTECT,
    )
    com_ca_total = models.DecimalField(
        "Chiffre d'affaire total HT",
        max_digits=10,
        decimal_places=0,
        null=True,
        editable=False,
        blank=True,
    )
    com_pct_ca_bio = models.DecimalField(
        "Part du CA bio (%)",
        max_digits=4,
        decimal_places=0,
        null=True,
        editable=False,
        blank=True,
    )
    # Accueil
    acc_act_tour = models.NullBooleanField("Activité touristique")
    acc_act_peda = models.NullBooleanField("Activité pédagogique")
    acc_rest_bio = models.NullBooleanField("Restauration bio")
    acc_act_autre = models.NullBooleanField("Autres activité")
    detail_activites = models.TextField(
        blank=True, verbose_name="détail des activités", default=''
    )
    # Lait
    lait_vaches = models.NullBooleanField("lait de vaches bio")
    lait_nb_vaches = models.PositiveIntegerField(
        "nombre de vaches laitières", null=True, blank=True
    )
    lait_quota_laiterie = models.PositiveIntegerField(
        "quota laiterie (en L)", null=True, blank=True
    )
    lait_quota_vente_directe = models.PositiveIntegerField(
        "quota vente directe (en L)", null=True, blank=True
    )
    lait_prod_laiterie = models.PositiveIntegerField(
        "Lait vendu en laiterie (en L)", null=True, blank=True
    )
    lait_prod_vente_directe = models.PositiveIntegerField(
        "Lait vendu en direct (en L)", null=True, blank=True
    )
    lait_laiterie = models.ForeignKey(
        ref.Laiterie,
        null=True,
        blank=True,
        verbose_name="laiterie",
        on_delete=models.PROTECT,
    )
    # viande issue du troupeau laitier
    lait_reformes_cl = models.PositiveIntegerField(
        "Nb d'animaux adultes type lait valorisés en circuit long (Réformes,"
        "Génisses, Boeufs)",
        null=True,
        blank=True,
    )
    lait_taurillons_cl = models.PositiveIntegerField(
        "Nb de veaux type lait valorisés en circuit long",
        null=True,
        blank=True,
    )
    lait_adultes_cc = models.PositiveIntegerField(
        "nb d'adultes laitiers vendus en circuit court / an",
        null=True,
        blank=True,
    )
    lait_veaux_cc = models.PositiveIntegerField(
        "nb de veaux laitiers vendus en circuit court / an",
        null=True,
        blank=True,
    )
    # Lait de chevre
    lait_chevres = models.NullBooleanField("chèvres bio")
    lait_nb_chevres = models.PositiveIntegerField(
        "nombre de chèvres", null=True, blank=True
    )
    lait_chevres_conversion = models.PositiveIntegerField(
        "nombre de chèvres en conversion", null=True, blank=True
    )
    lait_chevres_prod = models.PositiveIntegerField(
        "litres de lait de chèvre produits par an", null=True, blank=True
    )
    # Viande bovine
    bovine = models.NullBooleanField("viande bovine bio")
    bovine_nb_va = models.PositiveIntegerField(
        "nombre VA", null=True, blank=True
    )
    bovine_reformes_cl = models.PositiveIntegerField(
        "Nb d'animaux adultes type viande valorisés en circuit long",
        null=True,
        blank=True,
        help_text="(Réformes,Génisses, Boeufs)",
    )
    bovine_taurillons_cl = models.PositiveIntegerField(
        verbose_name="Nb de veaux type viande valorisés en circuit long",
        null=True,
        blank=True,
    )
    bovine_adultes_cc = models.PositiveIntegerField(
        "nb d'adultes race à viande vendus en circuit court / an",
        null=True,
        blank=True,
    )
    bovine_veaux_cc = models.PositiveIntegerField(
        "nb de veaux race à viande vendus en circuit court / an",
        null=True,
        blank=True,
    )
    # Porcs
    porcs = models.NullBooleanField("porc bio")
    porcs_nb_truies = models.PositiveIntegerField(
        "nombre de truies", null=True, blank=True
    )
    porcs_nb_com = models.PositiveIntegerField(
        "nombre de porcs commercialisés", null=True, blank=True
    )
    # Moutons
    moutons = models.NullBooleanField("moutons bio")
    moutons_nb_brebis = models.PositiveIntegerField(
        "nombre de brebis", null=True, blank=True
    )
    moutons_nb_agn_com = models.PositiveIntegerField(
        "nombre d'agneaux commercialisés", null=True, blank=True
    )
    # Petites animaux
    animaux_lapins = models.NullBooleanField("production de lapins bio")
    animaux_volailles = models.NullBooleanField(
        "production de volailles de chair bio"
    )
    animaux_volailles_chair = models.TextField(
        "Détail type de volailles (poulets, pintades, dindes)",
        blank=True,
        default='',
    )
    animaux_nb_poulets = models.PositiveIntegerField(
        "nb poulets (produits par an)", null=True, blank=True
    )
    animaux_nb_lapins = models.PositiveIntegerField(
        "nb lapins", null=True, blank=True
    )
    animaux_nb_autres_volailles = models.PositiveIntegerField(
        "nb autres volailles", null=True, blank=True
    )
    # Autres viandes
    animaux_autres_prod = models.NullBooleanField(
        "Autres productions animales bio"
    )
    animaux_details_autres_prod = models.TextField(
        "détail autres productions animale", blank=True, default=''
    )
    # Œufs
    oeufs_production_bio = models.NullBooleanField("production d'œufs bio")
    oeufs_nb_pondeuses = models.PositiveIntegerField(
        "nb de pondeuses", null=True, blank=True
    )
    oeufs_nb_oeufs = models.PositiveIntegerField(
        "nb d'œufs", null=True, blank=True
    )
    oeufs_acheteur_cl = models.ForeignKey(
        ref.AcheteurOeufs,
        null=True,
        blank=True,
        verbose_name="acheteur circuit long",
        on_delete=models.PROTECT,
    )
    # Miel
    miel_production_bio = models.NullBooleanField("apiculture bio")
    miel_nb_ruches = models.PositiveIntegerField(
        "nb de ruches", null=True, blank=True
    )
    # Formation
    formation_annee = models.NullBooleanField(
        "avez-vous participé à une formation au cours de l'année écoulée ?"
    )
    formation_annee_laquelle = models.ForeignKey(
        ref.Formation,
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name="si oui, laquelle ?",
    )
    formation_annee_commentaire = models.TextField(
        "commentaires", blank=True, default=''
    )
    formation_avenir = models.NullBooleanField(
        "avez-vous des besoins en formation à l'avenir ?"
    )
    formation_avenir_type = models.ForeignKey(
        ref.TypeFormation,
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name="si oui, lesquelles ?",
    )
    formation_avenir_details = models.TextField(
        "détails", blank=True, default=''
    )
    # Infos assolement
    ass_total = models.DecimalField(
        "Assolement total",
        blank=True,
        default=0,
        max_digits=10,
        decimal_places=4,
        help_text="Note: ce total n'est mis à jour "
        "qu'après enregistrement de la page",
    )
    ass_total_bio = models.DecimalField(
        "Assolement total bio",
        blank=True,
        default=0,
        max_digits=10,
        decimal_places=4,
        help_text=""
        "Note: ce total n'est mis à jour qu'après enregistrement de la page",
    )
    # block A garder pour usage futur 20150907
    lait_vaches_conversion = models.PositiveIntegerField(
        "nombre de vaches en conversion", null=True, blank=True
    )
    bovine_nb_conversion = models.PositiveIntegerField(
        "VA conversion", null=True, blank=True
    )
    # endblock A garder

    def __str__(self):
        "Retourne une représentation unicode de la structure."
        return "Données de %s pour l'année %s" % (self.structure, self.annee)

    def natural_key(self):
        return self.structure.natural_key() + [self.annee]

    def agri_id(self):
        return self.structure.id if self.structure else ""

    agri_id.short_description = "ID agriculteur bio"

    def get_absolute_url(self):
        "Retourne l'URL du formulaire de modification de l'objet."
        return reverse(
            'admin:%s_%s_change'
            % (self._meta.app_label, self.__class__.__name__.lower()),
            args=[self.pk],
        )

    def update_assolement(self, commit=False):
        """return change status, ass. total and ass. bio total,
        update field if update=True"""
        total = total_bio = 0
        for ass in self.assolement_set.all():
            sub_total, sub_total_bio = ass.get_total_sau()
            total += sub_total
            total_bio += sub_total_bio
        change = self.ass_total != total or self.ass_total_bio != total_bio
        if commit:
            self.ass_total = total
            self.ass_total_bio = total_bio
            self.save()
        return (change, total, total_bio)

    def agri_detail(self):
        if self.structure_id is None:
            return '-'
        return format_html(
            '<a href="{0}">{1}</a>',
            mark_safe(self.structure.get_absolute_url()),
            self.structure,
        )

    agri_detail.short_description = "Détail agriculteur"

    def clean(self):
        if self.lait_nb_vaches and self.lait_vaches is None:
            raise ValidationError(e_msg['production_vache_bio_non_choisi'])
        if self.lait_nb_chevres and self.lait_chevres is None:
            raise ValidationError(e_msg['production_chevre_bio_non_choisi'])
        if (
            self.bovine_nb_va
            or self.bovine_nb_conversion
            or self.bovine_reformes_cl
            or self.bovine_taurillons_cl
            or self.bovine_adultes_cc
            or self.bovine_veaux_cc
        ):
            if self.bovine is None:
                raise ValidationError(e_msg['production_bovin_bio_non_choisi'])
        if (self.porcs_nb_truies or self.porcs_nb_com) and self.porcs is None:
            raise ValidationError(e_msg['production_porc_bio_non_choisi'])
        if self.moutons_nb_brebis and self.moutons is None:
            raise ValidationError(e_msg['production_mouton_bio_non_choisi'])
        if self.animaux_nb_lapins and self.animaux_lapins is None:
            raise ValidationError(e_msg['production_lapin_bio_non_choisi'])
        if (
            self.animaux_nb_poulets or self.animaux_nb_autres_volailles
        ) and self.animaux_volailles is None:
            raise ValidationError(e_msg['production_volaille_bio_non_choisi'])
        if self.oeufs_nb_pondeuses and self.oeufs_production_bio is None:
            raise ValidationError(e_msg['production_oeuf_bio_non_choisi'])
        if (
            self.animaux_details_autres_prod
            and self.animaux_autres_prod is None
        ):
            raise ValidationError(e_msg['production_autre_bio_non_choisi'])
        if self.miel_nb_ruches and self.miel_production_bio is None:
            raise ValidationError(e_msg['production_miel_bio_non_choisi'])

    def get_messages(self, refresh=False):
        "fill info/warning messages at save"
        mlist = []
        if not self.com_ca_total:
            mlist.append((messages.WARNING, w_msg['commercialisation_vide']))
        return mlist


class Assolement(models.Model):
    "Informations d'assolement liées à DonneesBio (observatoire)."
    attrs_export = csv_assolement.attrs_export
    donneesbio = models.ForeignKey(
        'DonneesBio', verbose_name="objet donneesbio", on_delete=models.CASCADE
    )
    culture = models.ForeignKey(
        ref.Culture, verbose_name="culture", on_delete=models.PROTECT
    )
    detail_culture = models.CharField(
        "détail des cultures (si nécessaire)",
        max_length=255,
        default='',
        help_text="Complétez si nécessaire la description des cultures",
        blank=True,
    )
    surface_conv = models.DecimalField(
        "conventionnelle",
        null=True,
        blank=True,
        max_digits=10,
        decimal_places=4,
    )
    surface_c1 = models.DecimalField(
        "surface C1", null=True, blank=True, max_digits=10, decimal_places=4
    )
    surface_c2 = models.DecimalField(
        "surface C2", null=True, blank=True, max_digits=10, decimal_places=4
    )
    surface_c3 = models.DecimalField(
        "surface C3", null=True, blank=True, max_digits=10, decimal_places=4
    )
    surface_bio = models.DecimalField(
        "surface bio", null=True, blank=True, max_digits=10, decimal_places=4
    )
    tonnage = models.FloatField("tonnage", null=True, blank=True)
    client = models.ForeignKey(
        ref.ClientCulture,
        null=True,
        blank=True,
        verbose_name="client",
        on_delete=models.PROTECT,
    )
    prix = models.FloatField("Prix", null=True, blank=True)
    type_circuit = models.CharField(
        max_length=10, blank=True, choices=CHOIX_TYPE_CIRCUIT, default=''
    )

    def get_total_sau(self):
        "return sau and sau bio for this record"
        sub_total = sub_total_bio = 0
        for field in ('conv', 'c1', 'c2', 'c3', 'bio'):
            sub_total += getattr(self, 'surface_' + field, 0) or 0
            if field not in ('conv',):
                sub_total_bio += getattr(self, 'surface_' + field, 0) or 0
        return (sub_total, sub_total_bio)

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if not kwargs.pop('raw', False):
            self.donneesbio.update_assolement(commit=True)

    def delete(self, *args, **kwargs):
        super().delete(*args, **kwargs)
        if not kwargs.pop('raw', False):
            self.donneesbio.update_assolement(commit=True)

    def __str__(self):
        "representation unicode de l'objet"
        return "Assolement de %s" % self.culture


class Transformation(models.Model):
    "sous table des donnees bio"

    class Meta:
        verbose_name = "Transformation"
        verbose_name_plural = "Transformations"

    attrs_export = csv_transformation.attrs_export
    donneesbio = models.ForeignKey(
        'DonneesBio', verbose_name="objet donneesbio", on_delete=models.CASCADE
    )
    ferme = models.NullBooleanField("Transformation à la ferme")
    type_pduit = models.CharField(
        "Type de produit", max_length=20, choices=TYPE_TRANS, default=''
    )
    agrmt = models.CharField(
        "Agrément", max_length=20, choices=TYPE_AGREMENT, default=''
    )
    list_pduit = models.TextField(
        "Détail (liste des produits transformés)", blank=True, default=''
    )
    part_vol_glob = models.DecimalField(
        "Part du volume global",
        max_digits=10,
        decimal_places=0,
        null=True,
        blank=True,
        help_text="en pourcentage",
    )
    est_bio = models.NullBooleanField("Est bio ?")
    lait_abattoirs_vl = models.ManyToManyField(
        ref.Abattoir, blank=True, verbose_name="Abattoirs circuit court"
    )

    def __str__(self):
        return self.get_type_pduit_display() + " bio" if self.est_bio else ""


class Commercialisation(models.Model):
    "sous table des donnees bio"

    class Meta:
        verbose_name = "Commercialisation"
        verbose_name_plural = "Commercialisations"

    attrs_export = csv_commercialisation.attrs_export
    donneesbio = models.ForeignKey(
        'DonneesBio', verbose_name="objet donneesbio", on_delete=models.CASCADE
    )
    type_produit = models.CharField(
        "Type de produit",
        max_length=20,
        choices=TYPE_PRODUIT,
        default=TYPE_PRODUIT[0][0],
    )
    type_circuit = models.CharField(
        "Type de circuit",
        max_length=20,
        choices=TYPE_CIRCUIT,
        default=TYPE_CIRCUIT[0][0],
    )
    ch_aff = models.DecimalField(
        "Chiffre d'affaire HT",
        max_digits=10,
        decimal_places=0,
        blank=True,
        null=True,
    )
    detail = models.TextField(
        blank=True, help_text="horaire, date, lieux", default=''
    )
    bio = models.NullBooleanField("Est bio ?")

    def get_ch_aff(self):
        if self.ch_aff is None:
            return 0
        return self.ch_aff

    def get_total_ca_for_all_comm(self):
        "compute the total CA and the ratio of bio CA"
        ca_total = 0
        ca_bio = 0
        for comm in self.donneesbio.commercialisation_set.all():
            ca_total += comm.get_ch_aff()
            if comm.bio:
                ca_bio += comm.get_ch_aff()
        return (ca_total, (ca_bio * 100) // ca_total) if ca_total else (0, 0)

    def clean(self, *args, **kwargs):
        if self.bio is None:
            raise ValidationError(e_msg['commerce_bio_non_choisi'])
        return super().clean(*args, **kwargs)

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if not kwargs.pop('raw', False):
            (
                self.donneesbio.com_ca_total,
                self.donneesbio.com_pct_ca_bio,
            ) = self.get_total_ca_for_all_comm()
            self.donneesbio.save()

    def delete(self, *args, **kwargs):
        super().delete(*args, **kwargs)
        if not kwargs.pop('raw', False):
            (
                self.donneesbio.com_ca_total,
                self.donneesbio.com_pct_ca_bio,
            ) = self.get_total_ca_for_all_comm()
            self.donneesbio.save()


class Personne(models.Model):
    "Personne physique dépendant d'une structure."

    class Meta:
        ordering = ['nom', 'prenom', 'structure__designation']
        verbose_name_plural = "          Personne"
        indexes = [models.Index(fields=['nom'], name='contact_pers_nom_idx')]

    attrs_export = csv_personne.attrs_export
    attrs_publipostage = csv_personne.attrs_publipostage

    date_creation = models.DateField(
        "date de saisie de l'enregistrement", editable=False
    )
    date_derniere_modification = models.DateField(
        "date de dernière modification de informations", editable=False
    )
    utilisateur_derniere_modification = models.ForeignKey(
        auth_models.User,
        editable=False,
        on_delete=models.SET(get_rem_user),
        verbose_name="dernière modification par",
    )
    structure = models.ForeignKey(
        "Structure",
        verbose_name="affilié à",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    civilite = models.CharField(
        "civilité",
        max_length=10,
        blank=True,
        choices=CHOIX_CIVILITE,
        default='',
    )
    nom = models.CharField(max_length=255, default='')
    prenom = models.CharField(
        max_length=255, verbose_name="prénom", default=''
    )
    date_naissance = models.DateField(
        "date de naissance", null=True, blank=True
    )
    date_installation = models.DateField(
        "date d'installation", null=True, blank=True
    )
    fonction = models.CharField(
        "fonction", max_length=255, blank=True, default=''
    )
    adresse = models.TextField(
        "adresse personnelle",
        blank=True,
        default='',
        help_text="Saisissez l'adresse de la personne, sur 4 lignes maximum.",
    )
    etiquettes = models.ManyToManyField(
        ref.Etiquette, blank=True, verbose_name='étiquettes'
    )
    commune = models.ForeignKey(
        ref.Commune, null=True, blank=True, on_delete=models.PROTECT
    )
    # Certaines communes peuvent avoir plusieurs codes postaux
    code_postal = models.CharField(
        max_length=10,
        blank=True,
        default='',
        help_text="Laisser vide pour utiliser celui associé à la commune",
    )
    # la ville contient également le code cedex (à utiliser pour l'adresse)
    ville = models.CharField(
        max_length=255,
        blank=True,
        default='',
        help_text="Nom de ville"
        " avec code cédex si nécessaire, si différent du nom de la commune.",
    )
    courrier_structure = models.BooleanField(
        "Recevoir les courriers adressés à la structure",
        default=False,
        help_text="Cochez cette case si la personne doit recevoir les "
        "courriers papier envoyés à la structure.",
    )
    tel_fixe = models.CharField(
        "téléphone fixe", max_length=100, blank=True, default=''
    )
    tel_mobile = models.CharField(
        "téléphone mobile", max_length=100, blank=True, default=''
    )
    fax = models.CharField(max_length=100, blank=True, default='')
    courriel = models.CharField(max_length=255, blank=True, default='')
    pas_de_courriel = models.BooleanField(
        "pas d'envoi automatique de courriel",
        default=False,
        help_text="Cochez cette case si la personne ne souhaite"
        " pas recevoir de courriel envoyé automatiquement.",
    )
    reseaux_engagements = models.TextField(
        "Appartenance à des réseaux et engagements", blank=True, default=''
    )

    def __str__(self):
        """
        Retourne une représentation unicode de la structure.
        """
        if not self.nom and not self.prenom:
            return "Personne non nommée #%s" % self.id
        return "%s %s" % (self.prenom, self.nom)

    def get_absolute_url(self):
        """
        Retourne l'URl associée à l'objet.
        """
        return reverse(
            'admin:%s_%s_change'
            % (self._meta.app_label, self.__class__.__name__.lower()),
            args=[self.pk],
        )

    def nom_prenom(self):
        """
        Retourne les nom et prénom de la personne sous forme de chaîne de
        caractères.
        """
        return "%s %s" % (self.nom.upper(), self.prenom)

    def adresse_electronique(self):
        """
        Retourne une liste de 2 éléments:
        - le nom de la structure
        - l'adresses email associée
        Retourne une liste vide si aucune adresse courriel n'est renseignée.
        """
        if not self.courriel:
            return ()
        if self.nom:
            nom = "%s %s" % (self.prenom, self.nom)
        else:
            nom = ""
        return (nom, self.courriel)

    def get_courriels(self):
        """
        Méthode commune aux structures et personnes: retourne la liste des
        adresses emails (dans Personne il n'y en a qu'une)
        """
        adr = self.adresse_electronique()
        if adr and not self.pas_de_courriel:
            return [adr]
        return []

    def save(self, *args, **kwargs):
        """
        Remplit les champs date_creation et date_derniere_modification
        manuellement (la fonctionnalité auto_now de django est douteuse,
        mieux vaut le gérer manuellement).
        """
        if self.date_creation is None:
            self.date_creation = datetime.date.today()
        self.date_derniere_modification = datetime.date.today()
        super().save(*args, **kwargs)

    def get_messages(self, refresh=False):
        "fill info/warning messages at save"
        mlist = []
        if (
            self.structure_id
            and self.structure.statutbio > 0
            and not self.date_naissance
        ):
            mlist.append(
                (messages.WARNING, w_msg['pers_af_agri_sans_date_naiss'])
            )
        return mlist

    def liste_contacts_associes(self):
        return []

    def email_fb(self):
        return _field_fb_struct(self, 'courriel')

    email_fb.short_description = "courriel (pers>pm)"

    def tel_fixe_fb(self):
        return _field_fb_struct(self, 'tel_fixe')

    tel_fixe_fb.short_description = "tel fixe (pers>pm)"

    def tel_mobile_fb(self):
        return _field_fb_struct(self, 'tel_mobile')

    tel_mobile_fb.short_description = "tel mob (pers>pm)"

    def adr_adresse_fb(self):
        return _field_fb_struct(self, 'adresse')

    adr_adresse_fb.short_description = "adresse (pers>pm)"

    def adr_ville_fb(self):
        return _field_fb_struct(self, 'ville', 'adresse')

    adr_ville_fb.short_description = "ville (pers>pm)"

    def adr_cp_fb(self):
        return _field_fb_struct(self, 'code_postal', 'adresse')

    adr_cp_fb.short_description = "code postal (pers>pm)"


# ## Classes secondaires
class RdvPersonne(models.Model):
    "modele intermediaire: participation d'une personne a un evt"

    class Meta:
        db_table = "contacts_rdv_personnes"
        unique_together = ('rdv', 'personne')

    attrs_export = csv_rdvpersonne.attrs_export
    rdv = models.ForeignKey('Rdv', on_delete=models.CASCADE)
    personne = models.ForeignKey(
        'Personne',
        verbose_name="Personne physique",
        help_text="pour les rendez vous collectifs",
        on_delete=models.CASCADE,
    )

    def is_organisateur(self):
        if self.personne.structure:
            return format_html(
                '<a href="{0}">{1}</a>',
                mark_safe(self.personne.structure.get_absolute_url()),
                self.personne.structure,
            )
        return ''

    is_organisateur.short_description = "Organisation"

    def __str__(self):
        return "Participation de {0} à {1}".format(self.personne, self.rdv)


class RdvPersonneMorale(models.Model):
    "modele intermediaire: participation d'une structure a un rdv"

    class Meta:
        db_table = "contacts_rdv_personnes_morales"
        unique_together = ('rdv', 'structure')

    rdv = models.ForeignKey('Rdv', on_delete=models.CASCADE)
    structure = models.ForeignKey(
        'Structure',
        verbose_name="Personne morale",
        help_text="pour les rendez vous collectifs",
        on_delete=models.CASCADE,
    )

    def __str__(self):
        return "Participation de {0} à {1}".format(self.structure, self.rdv)


class Rdv(models.Model):
    "Rendez vous generiques"

    class Meta:
        verbose_name = "Rendez vous ou Évenement"
        verbose_name_plural = "Rendez vous et Évenements"
        ordering = ['-date']
        get_latest_by = 'date'

    attrs_rdv_coll = csv_rdv.attrs_rdv_coll
    attrs_export = csv_rdv.attrs_export  # rdv ind
    date = models.DateField(db_index=True)
    technicien = models.ForeignKey(
        auth_models.User, on_delete=models.SET(get_rem_user)
    )
    type_rdv = models.CharField(
        max_length=25, choices=CHOIX_TYPE_RDV, verbose_name="type", default=''
    )
    projet_bio = models.ForeignKey(
        'ProjetBio',
        blank=True,
        null=True,
        verbose_name="Projet bio",
        on_delete=models.SET_NULL,
    )
    titre_evt = models.ForeignKey(
        'references.Formation',
        blank=True,
        null=True,
        verbose_name="Désignation évenement",
        on_delete=models.SET_NULL,
    )
    personne_morale = models.ForeignKey(
        'Structure',
        related_name='rdv_ind_set',
        blank=True,
        null=True,
        verbose_name="Personne morale",
        help_text="pour les rendez vous individuels",
        on_delete=models.CASCADE,
    )
    personnes_morales = models.ManyToManyField(
        'Structure',
        related_name='evt_set',
        through="RdvPersonneMorale",
        verbose_name="Participants",
        help_text="pour les rendez vous collectifs",
    )
    personnes = models.ManyToManyField(
        'Personne', through='RdvPersonne', verbose_name="Participant(s) ind."
    )
    territoire_cible = models.ManyToManyField(
        'references.Territoire', blank=True
    )
    public_cible = models.ManyToManyField('references.PublicCible', blank=True)
    organisateur = models.ForeignKey(
        Personne,
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        related_name='organisation_pers',
    )
    intervenant = models.ForeignKey(
        Personne,
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        related_name='intervenant_pers',
    )
    objectif = models.TextField(blank=True, default='')
    res_attendu = models.TextField(blank=True, default='')
    lieu = models.ForeignKey(
        ref.Commune,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        help_text="si vide, rempli automatiquement",
    )
    # accounting: duree et distance associees au rendez vous
    acc_duree = models.DecimalField(
        "Temps passé",
        max_digits=3,
        decimal_places=1,
        null=True,
        blank=True,
        help_text="en heure",
    )
    acc_distance = models.DecimalField(
        "Distance associée",
        max_digits=4,
        decimal_places=1,
        null=True,
        blank=True,
        help_text="en km",
    )
    finance_prog = models.ForeignKey(
        ref.FinanceProg,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Financement / Programme concerné",
    )
    est_confirme = models.BooleanField(
        "est confirmé", blank=True, default=False
    )
    commentaire = models.TextField(blank=True, default='')
    pk_rdv_suivi = models.IntegerField(default=0, editable=False)
    pk_rdv_bio = models.IntegerField(default=0, editable=False)

    def struct_af_item(self, item):
        if (
            getattr(self, item) is None
            or getattr(self, item).structure is None
        ):
            return 'N/A'
        return format_html(
            '<a href="{0}" target="_BLANK">{1}</a>',
            mark_safe(getattr(self, item).structure.get_absolute_url()),
            getattr(self, item).structure,
        )

    def struct_af_org(self):
        return self.struct_af_item('organisateur')

    struct_af_org.short_description = "P. morale affiliée à l'organisation"

    def struct_af_int(self):
        return self.struct_af_item('intervenant')

    struct_af_int.short_description = "P. morale affiliée à l'intervenant"

    def proj_id(self):
        "retoune l'identifiant du projet"
        return self.projet_bio.id if self.projet_bio else ""

    proj_id.short_description = "ID Projet"

    def get_cat(self):
        "retourne la categorie de rendez vous (indiv, collect)"
        resp = set(
            key
            for key, value in MTX_TYPE_RDV.items()
            if self.type_rdv in value
        )
        return resp

    def get_absolute_url(self):
        return reverse(
            'admin:{0}_{1}_change'.format(
                self._meta.app_label, self.__class__.__name__.lower()
            ),
            args=[self.pk],
        )

    def save(self, *args, **kwargs):
        if self.lieu is None:
            pers_morale = self.rdvpersonnemorale_set.all()[:1]
            if len(pers_morale) == 1:
                pers_morale = pers_morale[0].structure
                self.lieu = pers_morale.commune
        super().save(*args, **kwargs)

    def clean(self):
        cat = self.get_cat()
        if self.est_confirme:
            if 'pbio' in cat and not self.projet_bio:
                raise ValidationError(e_msg['rdv_pbio_sans_pbio'])
            if 'indiv' in cat and not self.personne_morale:
                raise ValidationError(e_msg['rdv_indiv_sans_part'])
            if 'collect' in cat:
                if not self.titre_evt:
                    raise ValidationError(e_msg['rdv_collec_sans_titre'])

    def __str__(self):
        return "{0}-{1}-{2}".format(
            self.date.strftime(LOCAL_DATE_FORMAT),
            self.get_type_rdv_display(),
            self.technicien,
        )

    def get_messages(self):
        ""
        mlist = []
        if (
            'collect' in self.get_cat()
            and self.est_confirme
            and not self.rdvpersonne_set.exists()
        ):
            mlist.append((messages.WARNING, w_msg['rdv_collec_sans_part']))
        return mlist


class AdhesionManager(models.Manager):
    def get_by_natural_key(self, structure_nom, annee):
        return self.get(structure__designation=structure_nom, annee=annee)


class Adhesion(models.Model):
    "Adhésion"

    class Meta:
        ordering = ['-date']
        get_latest_by = 'date'
        verbose_name = "adhésion GAB"
        verbose_name_plural = "      Adhésions GAB"
        unique_together = [['structure', 'annee']]

    objects = AdhesionManager()

    # Champs à inclure dans un export CSV complet
    attrs_export = ['structure', 'date', 'montant', 'mode_reg']

    structure = models.ForeignKey(
        'Structure', verbose_name="Personne morale", on_delete=models.CASCADE
    )
    date = models.DateField(db_index=True)
    montant = models.DecimalField(
        "montant de l'adhésion (en euros)", max_digits=8, decimal_places=2
    )
    annee = models.IntegerField(
        "Adhésion de l'année",
        null=True,
        blank=True,
        help_text="Laisser vide pour utiliser l'année de la date d'adhésion",
    )
    mode_reg = models.CharField(
        "Mode de règlement",
        max_length=16,
        blank=True,
        choices=MODE_REGLT,
        default='',
    )

    def __str__(self):
        """
        Représentation unicode de l'objet (utilisé notamment pour affichage
        dans l'objet Structure).
        """
        return "Adhésion[%i] du %s de %.2f EUR de %s" % (
            self.annee,
            self.date.strftime(LOCAL_DATE_FORMAT),
            self.montant,
            self.structure,
        )

    def clean(self):
        if self.annee is None:
            self.annee = self.date.year

    def save(self, *args, **kwargs):
        """
        Surchargé pour mettre à jour les dates d'adhésions de l'agriculteur.
        """
        super().save(*args, **kwargs)
        self._update_dates_structure()

    def delete(self, *args, **kwargs):
        """
        Surchargé pour mettre à jour les dates d'adhésions de l'agriculteur.
        (cette méthode n'est pas appelée pour les suppressions en masse,
        mais ça n'arrivera pas dans un usage normal de l'application).
        """
        super().delete(*args, **kwargs)
        self._update_dates_structure()

    def _update_dates_structure(self):
        """
        Calcule et sauvegarde les dates de première et dernière adhésion,
        repectivement date_adhesion et date_renouvellement dans l'objet
        Structure
        """
        adhesions_qs = Adhesion.objects.filter(
            structure=self.structure
        ).order_by('date')
        if len(adhesions_qs) != 0:
            premiere_adhesion = adhesions_qs[0]
        else:
            premiere_adhesion = None
        if len(adhesions_qs) > 1:
            # le renouvellement est la dernière adhésion en date
            renouvellemenent = adhesions_qs[len(adhesions_qs) - 1]
        else:
            renouvellemenent = None
        self.structure.premiere_adhesion = premiere_adhesion
        self.structure.renouvellement_adhesion = renouvellemenent
        self.structure.save()


class DocumentJoint(models.Model):
    "Documents joints aux rendez-vous."

    class Meta:
        verbose_name = "document joint"
        verbose_name_plural = "Documents Joints"

    document = models.FileField(upload_to='pj/%Y/%m/%d/')
    date_depot = models.DateTimeField(auto_now=True)
    rdv = models.ForeignKey(
        Rdv, null=True, blank=True, on_delete=models.CASCADE
    )

    def __str__(self):
        "Retourne une représentation unicode de l'objet."
        return "%s : %s" % (self.document, self.date_depot)
