# -*- coding: utf-8 -*-
"""
walk throuht django Model Class
"""

import re

from django.apps import apps
from django.contrib.admin.utils import label_for_field
from django.db import models
from django.db.models.fields.related import (
    ForwardManyToOneDescriptor,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
)

AGR_FUNC = {'count': '%s (nb)', 'sum': '%s (somme)'}

RE_GET_DISPLAY = re.compile(r'get_(.*)_display')


def get_related_model(model, field_name):
    """
    return model referenced by field_name
    if field_name does not reference another model, return None
    """
    pttt = getattr(model, field_name, None)
    if isinstance(
        pttt, (ForwardManyToOneDescriptor, ReverseManyToOneDescriptor)
    ):
        return pttt.field.related_model
    if isinstance(pttt, (ReverseManyToOneDescriptor, ManyToManyDescriptor)):
        return pttt.related.model
    if callable(pttt):
        target_model = getattr(pttt, 'target_model', None)
        if target_model is not None:
            model_args = target_model.split('.', 1)
            if len(model_args) == 1:
                model_args = [model._meta.app_label, model_args[0]]
            return apps.get_model(*model_args)
    return None


def label_for_composed_field(model, composed_name):
    """
    expect a chained named in django manner:
    name1__name2__function
    """
    composed_name = composed_name.split('__')
    composed_name_len = len(composed_name)

    if composed_name_len == 0:
        return None

    cur_model = next_model = model
    model_list = [model]
    for idx, name in enumerate(composed_name):
        cur_model = next_model
        next_model = get_related_model(cur_model, name)
        if next_model:
            model_list.append(next_model)
        else:
            break

    if composed_name_len > 1 and name in AGR_FUNC:
        label = label_for_field(composed_name[idx - 1], model_list[idx - 1])
        label = AGR_FUNC[name] % label
        return label

    match_get_display = RE_GET_DISPLAY.match(name)
    if match_get_display:
        name = match_get_display.groups()[0]

    label = label_for_field(name, cur_model)
    return label


def get_fk_name_from_model(model):
    "take an object or a model and return the list of ForeignKey field name"
    return [
        elm.name
        for elm in model._meta.fields
        if isinstance(elm, models.ForeignKey)
    ]


def get_m2m_name_from_model(model):
    "take an object or a model and return list of ManyToMany field name"
    return [field.name for field in model._meta.many_to_many]


# EOF
